"""
    Copyright (C) 2008  Sreejith K <sreejithemk@gmail.com>
    
    This file is part of PyDCS. 

    This program (PyDCS) is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program (PyDCS) is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Author : Sreejith K
    Email : sreejithemk@gmail.com
    Web : http://sreejithemk.net
"""



#  Define classes for Connector, Logic Circuit, Gate


class Connector :
    def __init__ (self, owner, name, activates=0, monitor=0,connected=0) :
        self.value = None
        self.owner = owner
        self.name  = name
        self.monitor  = monitor
        self.connects = []
        self.activates = activates 
	self.connected = connected  

    def connect (self, inputs) :
        if type(inputs) != type([]) : inputs = [inputs]
        for input in inputs : self.connects.append(input)


    def set (self, value) :
        if self.value == value : return      
        self.value = value
        if self.activates : self.owner.evaluate()
        if self.monitor :
            print "Connector %s-%s set to %s" % (self.owner.name,self.name,self.value)
        for con in self.connects : con.set(value)

class LC :
    def __init__ (self, name) :
        self.name = name
    def evaluate (self) : return
            
class Not (LC) :         
    def __init__ (self, name) :
        LC.__init__ (self, name)
        self.A = Connector(self,'A', activates=1)
        self.B = Connector(self,'B',monitor=1)
    def evaluate (self) : self.B.set(not self.A.value)

class Node (LC) :
    def __init__ (self , name):
        LC.__init__(self,name)
	self.A = Connector(self, 'A',activates=1)
	self.B = Connector(self, 'B',monitor=1)
    def evaluate (self) : self.B.set(self.A.value)

class Gate2 (LC) :         
    def __init__ (self, name) :
        LC.__init__ (self, name)
        self.A = Connector(self,'A',activates=1)
        self.B = Connector(self,'B',activates=1)
        self.C = Connector(self,'C', monitor=1)

class And (Gate2) :       
    def __init__ (self, name) :
        Gate2.__init__ (self, name)
    def evaluate (self) : self.C.set(self.A.value and self.B.value)

class Or (Gate2) :         
    def __init__ (self, name) :
        Gate2.__init__ (self, name)
    def evaluate (self) : self.C.set(self.A.value or self.B.value)

class Xor (Gate2) :
    def __init__ (self, name) :
        Gate2.__init__ (self, name)
        self.A1 = And("A1") 
        self.A2 = And("A2")
        self.I1 = Not("I1")
        self.I2 = Not("I2")
        self.O1 = Or ("O1")
        self.A.connect    ([ self.A1.A, self.I2.A])
        self.B.connect    ([ self.I1.A, self.A2.A])
        self.I1.B.connect ([ self.A1.B ])
        self.I2.B.connect ([ self.A2.B ])
        self.A1.C.connect ([ self.O1.A ])
        self.A2.C.connect ([ self.O1.B ])
        self.O1.C.connect ([ self.C ])


class HalfAdder (LC) :         
    def __init__ (self, name) :
        LC.__init__ (self, name)
	self.A = Connector(self,'A',1)
        self.B = Connector(self,'B',1)
        self.S = Connector(self,'S',monitor=1)
        self.Cout = Connector(self,'Cout',monitor=1)
        self.X1= Xor("X1")
        self.A1= And("A1")
        self.A.connect    ([ self.X1.A, self.A1.A])
        self.B.connect    ([ self.X1.B, self.A1.B])
        self.X1.C.connect ([ self.S])
        self.A1.C.connect ([ self.Cout])

class FullAdder (LC) :         
    def __init__ (self, name) :
        LC.__init__ (self, name)
	self.A    = Connector(self,'A',1)
        self.B    = Connector(self,'B',1)
        self.Cin  = Connector(self,'Cin',1)
        self.S    = Connector(self,'S',monitor=1)
        self.Cout = Connector(self,'Cout',monitor=1)
        self.H1= HalfAdder("H1")
        self.H2= HalfAdder("H2")
        self.O1= Or("O1")
        self.A.connect    ([ self.H1.A ])
        self.B.connect    ([ self.H1.B ])
        self.Cin.connect  ([ self.H2.A ])
        self.H1.S.connect ([ self.H2.B ])
        self.H1.Cout.connect ([ self.O1.B])
        self.H2.Cout.connect ([ self.O1.A])
        self.H2.S.connect ([ self.S])
        self.O1.C.connect ([ self.Cout])

def bit (x, bit) : return x[bit]=='1' 

def test4Bit (a, b) :    
    F0 = FullAdder ("F0")
    F1 = FullAdder ("F1"); F0.Cout.connect(F1.Cin)
    F2 = FullAdder ("F2"); F1.Cout.connect(F2.Cin)
    F3 = FullAdder ("F3"); F2.Cout.connect(F3.Cin)
    
    F0.Cin.set(0)
    F0.A.set(bit(a,3)); F0.B.set(bit(b,3))  
    F1.A.set(bit(a,2)); F1.B.set(bit(b,2))
    F2.A.set(bit(a,1)); F2.B.set(bit(b,1))
    F3.A.set(bit(a,0)); F3.B.set(bit(b,0))

    print F3.Cout.value,
    print F3.S.value,
    print F2.S.value,
    print F1.S.value,
    print F0.S.value,

def testFull (a,b,c) :
    F1 = FullAdder ("F1")
    F1.Cin.set(c)
    F1.A.set(a)
    F1.B.set(b)
    
    print "Cin=%d  A=%d  B=%d" % (c,a,b)
    print "Sum=%d  Cout=%d" % (F1.S.value, F1.Cout.value)
